

class BaseCssSelect(object):
    """base class cssselect fields"""
    element_method = None
    extra_data = False
    attr = False
    text = False
    text_content = False

    def __init__(self, add_domain=False, save_start_url=False, save_url=False, many=False, *args, **kwargs):
        self.add_domain = add_domain
        self.save_start_url = save_start_url
        self.save_url = save_url
        self.many = many
        self.attr_name = None
        self.attr_data = None

    def __set__(self, instance, value):
        value = self._check_to_field_type(instance, value)
        instance.__dict__[self.attr_name] = value

    def _except_attr_type_error(self, attr, instance=None):
        raise TypeError('attribute {} of the object {} must be a type str in {}'.format(attr, self, instance))

    def _check_to_field_type(self, instance, value):
        if not isinstance(value, str):
            self._except_attr_type_error(self.attr_name, instance)
        return value


class ExtraDataField(BaseCssSelect):
    extra_data = True


class TextCssSelect(BaseCssSelect):
    element_method = '.text'
    text =True


class TextContentCssSelect(BaseCssSelect):
    element_method = '.text_content()'
    text_content = True


class BodyCssSelect(BaseCssSelect):
    body = True

    def __init__(self, start_url=None, body_count=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start_url = start_url
        self.body_count = body_count


class AttrCssSelect(BaseCssSelect):
    attr = True
    NAME_ATTRIBUTE = 'attr_data'
    element_method = '.get("{}")'

    def __init__(self, attr_data=None, *args, **kwargs):
        attr_data = self._check_attr_data_is_required(attr_data=attr_data, *args, **kwargs)
        super().__init__(*args, **kwargs)
        self.attr_data = attr_data
        self.element_method = self.element_method.format(attr_data)

    def _check_attr_data_is_required(self, *args, **kwargs):
        """check the required attribute attr_data and type"""
        kwargs_attr = kwargs.get(self.NAME_ATTRIBUTE, False)
        if not kwargs_attr and not args:
            raise AttributeError('object {} has a required attribute {}'.format(self, self.NAME_ATTRIBUTE))
        if kwargs_attr and not isinstance(kwargs_attr, str) or args and not isinstance(args[0], str):
            self._except_attr_type_error(self.NAME_ATTRIBUTE)
        return kwargs_attr if kwargs_attr else args[0]


class ImgCssSelect(AttrCssSelect):
    img = True

    def __init__(self, *args, **kwargs):
        def_kwargs = self.__set_default_attr(*args, **kwargs)
        super().__init__(*args, **def_kwargs)

    def __set_default_attr(self, *args, **kwargs) -> dict:
        """Set the default attr data attribute to 'src' for the ImgCSSSelect field"""
        try:
            if self._check_attr_data_is_required(*args, **kwargs):
                return kwargs
        except AttributeError:
            kwargs.update({self.NAME_ATTRIBUTE: 'src'})
            return kwargs






from djparsing.data import BaseCssSelect


class ParserExceptions(Exception):
    default_detail = 'A parser error'
    default_field = 'system_error'
    default_obj = 'Parser'

    @staticmethod
    def _get_error(detail, field=None, obj=None):
        return '{}: {} in {}'.format(field, detail, obj)

    def __init__(self, detail=None, field=None, obj=None):
        if detail is None:
            detail = self.default_detail
        if field is None:
            field = self.default_field
        if obj is None:
            obj = self.default_obj

        self.detail = self._get_error(detail, field, obj)

    def __str__(self):
        return self.detail


class URLException(ParserExceptions):
    default_detail = 'Check attribute'
    default_obj = None
    default_field = 'url'

    @staticmethod
    def _get_error(detail, field=None, obj=None):
        return '{}: {} in {}'.format(detail, field, obj)


class FieldException(ParserExceptions):
    default_detail = 'Check the initialization '
    default_field = 'BaseCssSelect'
    default_obj = BaseCssSelect()





from time import sleep, strftime
from djparsing.core import Parser
from djparsing.data import BodyCssSelect, TextContentCssSelect, AttrCssSelect, TextCssSelect, ImgCssSelect, \
    ExtraDataField

print(strftime('[%H: %M: %S]'), end=' ')
print('********************START*********************')


class HabrBlog(Parser):
    b = BodyCssSelect(body_count=13)
    text = TextContentCssSelect()
    source = AttrCssSelect(attr_data='href')
    title = TextCssSelect()
    img = ImgCssSelect('src')


data_habr_blog = {
    'b': 'div.posts_list > ul.content-list > li.content-list__item > article.post_preview',
    'text': 'div.post__body_crop > div.post__text',
    'source': 'h2.post__title > a',
    'title': 'h2.post__title > a',
    'img': '.post__body_crop > .post__text img',
}


LIST_WORDS = ['Python', 'Django', 'Питон', 'Flask', 'PyDev', 'sanic', 'aiohttp', '2018', 'vue', 'computer']


class HabrAll(HabrBlog):

    class Meta:
        coincidence = LIST_WORDS
        field_coincidence = 'title'


obj_habr_all = HabrAll(
    url='https://habrahabr.ru',
    **data_habr_blog
)


class RealPython(HabrBlog):
    source = AttrCssSelect(attr_data='href', add_domain=True)


obj_realpython = RealPython(
    b="div.col-12.mb-3",
    text="div.card > div.card-body > p.my-1",
    source="div.card > a",
    title="div.card > div.card-body > a > h2.card-title",
    img="div.card > a > img.card-img-top",
    url='https://realpython.com/blog/'

)


class PythonzPars(Parser):
    start = BodyCssSelect(start_url='div.listing_item > div.description > a', add_domain=True, body_count=45)
    source = AttrCssSelect(attr_data="src")
    title = AttrCssSelect("data-geopattern")
    url = ExtraDataField(save_start_url=True)


obj_pthonz = PythonzPars(
    start='div.col-md-9',
    source="div.col-md-9 > div > iframe",
    title="div > h1",
    url='http://pythonz.net/videos/'
)


class ItProgerPars(Parser):
    source = ExtraDataField(save_start_url=True)
    body = BodyCssSelect(
        start_url='article > div.entry-image a',
        body_count=3)
    img = ImgCssSelect('data-lazy-src')
    text = TextContentCssSelect()
    title = TextCssSelect()

    class Meta:
        coincidence = LIST_WORDS
        field_coincidence = 'title'


obj_itproger = ItProgerPars(
    body="div.container div#content > article.post",
    text="div.entry-container > div.entry-content > p",
    source=".post-title .title-box .entry-title a",
    title=".post-title > h1.entry-title",
    img="div.entry-image > a > img",
    url='https://tproger.ru'
)


if __name__ == '__main__':
    obj_itproger.run(log=True)
    obj_pthonz.run(log=True)
    print(strftime('[%H: %M: %S]'), end=' ')
    print('********************END*********************')





import hashlib
from urllib.parse import urljoin

from .exceptions import FieldException


class ResultParser(object):
    def __init__(self, obj_parser):
        self.obj_parser = obj_parser

    def __repr__(self):
        return 'ResultParser({})'.format(self.obj_parser)

    def get_object_field(self, field_name):
        return self.obj_parser.__class__.__dict__[field_name]

    def get_elem_result(self, block, obj_field):
        elem = block.cssselect(self.obj_parser.__getattribute__(obj_field.attr_name))[0]
        if obj_field.text:
            return elem.text
        elif obj_field.text_content:
            return elem.text_content()
        elif obj_field.attr:
            return elem.get(obj_field.attr_data)

    def get_field(self, block):
        field = self.obj_parser.get_field_coincidence()
        obj = self.get_object_field(field)
        try:
            res_field = self.get_elem_result(block=block, obj_field=obj)
        except IndexError:
            raise FieldException(field=field, obj=obj)
        else:
            return res_field

    def is_word_in_field(self, field):
        list_coincidence = self.obj_parser.get_list_coincidence()
        for word in list_coincidence:
            if word.lower() in field.lower():
                return True
        return False

    def __iter__(self):
        for ind, block in enumerate(self.obj_parser.block_list):
            yield ind
            if self.obj_parser.is_field_coincidence():
                field_result = self.get_field(block)
                if not self.is_word_in_field(field=field_result):
                    continue
            for attr_name in self.obj_parser.cls_attr:
                obj = self.get_object_field(attr_name)
                if obj.save_start_url and hasattr(obj, 'extra_data'):
                    try:
                        yield {attr_name: self.obj_parser.get_page_url(ind=ind)}
                    except IndexError:
                        raise ValueError('Check the field with the attribute "start_url"')
                    continue
                if obj.save_url:
                    yield {attr_name: self.obj_parser.url}
                    continue
                if obj.many:
                    data_input = list()
                    for data in block.cssselect(self.obj_parser.__getattribute__(attr_name)):
                        if obj.attr:
                            data_input.append(data.get(obj.attr_data))
                        if obj.text:
                            data_input.append(data.text)
                    yield {attr_name: data_input}
                    continue
                try:
                    pars_result = self.get_elem_result(block=block, obj_field=obj)
                except IndexError:
                    if attr_name == self.obj_parser.get_field_image():
                        continue
                    else:
                        raise FieldException(field=attr_name, obj=self.obj_parser)
                if attr_name in self.obj_parser.get_fields_add_domain():
                    pars_result = urljoin(self.obj_parser.get_base_domain(), pars_result)
                if attr_name == self.obj_parser.get_field_image() and pars_result:
                    pars_result = self.obj_parser.uploaded_image(
                        pars_result,
                        '{}.jpg'.format(hashlib.sha1(pars_result.encode('utf-8')).hexdigest())
                    )
                yield {attr_name: pars_result}




from urllib.parse import urlsplit


class Options(object):
    def __init__(self, meta):
        self.meta = meta
        self.list_domain = list()
        self.image = None
        self.base_domain = None
        self.page_url = list()
        self.data_result = dict()

    @staticmethod
    def get_domain(url):
        return '{0.scheme}://{0.netloc}'.format(urlsplit(url))



import copy
# import hashlib
import os
import logging
# from abc import ABCMeta
import shutil
from io import BytesIO
from urllib.error import URLError
import lxml
import requests
from PIL import Image
from lxml.html import fromstring
# from memory_profiler import profile
from requests.exceptions import MissingSchema
from .exceptions import URLException, FieldException
from .parser import ResultParser
from .options import Options
from .data import BaseCssSelect
from .settings import PATH_TEMP
from django.core.files.uploadedfile import InMemoryUploadedFile
# from django.core.files.uploadedfile import TemporaryUploadedFile
from django.apps import apps


def init(**kwargs):
    def dec(cls):
        if hasattr(cls, 'Meta'):
            meta = cls.Meta
            if not hasattr(meta, 'model'):
                meta.model = apps.get_model(kwargs['app'], kwargs['model'])
        else:
            setattr(cls, 'Meta', type('Meta', (), {'model': apps.get_model(kwargs['app'], kwargs['model'])}))
        return cls

    return dec


class ParserMeta(type):
    # meta class for classes parsing
    def __new__(cls, name, bases, attrs):

        parents = [base for base in bases if isinstance(base, ParserMeta)]
        if not parents:
            return super().__new__(cls, name, bases, attrs)

        new_cls = super().__new__(cls, name, bases, attrs)
        for key, attr in new_cls.__dict__.items():
            if isinstance(attr, (BaseCssSelect,)):
                attr.attr_name = key

        # we copy from the base classes fields and class Meta
        for base in reversed(parents):
            if not attrs.get('Meta', None):
                if hasattr(base, 'Meta'):
                    base_meta = copy.deepcopy(base.Meta)
                    setattr(new_cls, 'Meta', base_meta)
            for key, attr in base.__dict__.items():
                if isinstance(attr, BaseCssSelect):
                    if key not in new_cls.__dict__:
                        parent_fields = copy.deepcopy(key)
                        setattr(new_cls, parent_fields, attr)
        meta = getattr(new_cls, 'Meta', None)
        setattr(new_cls, '_opt', Options(meta))

        return new_cls


class Parser(object, metaclass=ParserMeta):
    __slots__ = ('url', 'add_field', 'block_list', 'page_url', '_opt', 'cls_attr', 'data_db')

    def __init__(self, url=None, **kwargs):
        self.url = url
        self.data_db = None
        self.cls_attr = set()
        self.add_field = dict()
        for key, value in kwargs.items():
            setattr(self, key, value)
        for key, attr in self.__class__.__dict__.items():
            if isinstance(attr, BaseCssSelect):
                if attr.add_domain:
                    self._opt.base_domain = self._opt.get_domain(self.url)
                    self._opt.list_domain.append(key)
                if hasattr(attr, 'body'):
                    if attr.start_url:
                        self.block_list = self._get_block_html(key, attr, attr.body_count, start_url=True, **kwargs)
                    else:
                        self.block_list = self._get_block_html(key, attr, attr.body_count, **kwargs)
                    continue
                if hasattr(attr, 'img'):
                    self._opt.image = key
                self.cls_attr.add(key)
        if not hasattr(self, 'block_list'):
            raise FieldException(detail='The required field', field='BodyCssSelect', obj=self.__class__)

    def __str__(self):
        return self.__class__.__name__

    def get_page_url(self, ind):
        return self._opt.page_url[ind]

    def get_field_image(self):
        return self._opt.image

    def get_fields_add_domain(self):
        return self._opt.list_domain

    def get_base_domain(self):
        return self._opt.base_domain

    def get_field_coincidence(self):
        return self._opt.meta.field_coincidence

    def get_element_method(self, attr_model):
        return self.__class__.__dict__[attr_model].element_method

    def get_list_coincidence(self):
        return getattr(self.__class__.Meta, 'coincidence')

    def _gen_block_html(self, urls, key, **kwargs):
        for url in urls:
            yield self._get_html(url=url, **kwargs).cssselect(self.__getattribute__(key))[0]

    def _get_block_html(self, key, attr, body_count, start_url=False, **kwargs):
        count = body_count if body_count else 30
        try:
            if start_url:
                for elem_url in self._get_html(**kwargs).cssselect(attr.start_url)[0:count]:

                    if self._opt.base_domain:
                        self._opt.page_url.append('{0}{1}'.format(self._opt.base_domain, elem_url.get("href")))
                    else:
                        self._opt.page_url.append(elem_url.get("href"))
                return list(self._gen_block_html(self._opt.page_url, key, **kwargs))
            return self._get_html(**kwargs).cssselect(self.__getattribute__(key))[0:count]
        except IndexError:
            raise FieldException(field=key, obj=self)
        except MissingSchema:
            raise URLException(obj=self)

    def _get_html(self, url=None, web_driver=False, **kwargs) -> lxml.html.HtmlElement:
        if url is None:
            url = self.url
        if web_driver:
            from selenium import webdriver
            driver = webdriver.Firefox()
            driver.get(url)
            response = driver.page_source
            driver.close()
        else:
            response = requests.get(url).text
        return fromstring(response)

    @classmethod
    def set_obj(cls, url, flag):
        obj = cls(url=url)
        obj.add_field[flag] = True
        return obj

    @staticmethod
    def uploaded_image(url, name):
        try:
            os.chdir(PATH_TEMP)
        except FileNotFoundError:
            os.makedirs(PATH_TEMP)
            os.chdir(PATH_TEMP)
        finally:
            try:
                resp_img = requests.get(url, stream=True)
                if resp_img.status_code == 200:
                    with open(name, 'wb') as img:
                        resp_img.raw.decode_content = True
                        shutil.copyfileobj(resp_img.raw, img)
                # urlretrieve(url, name)
            except URLError:
                return None
            image = Image.open(name)
            image_io = BytesIO()
            image.save(image_io, "png", optimize=True)
            image_io.seek(0)
            os.remove(os.path.join(PATH_TEMP, name))
        return InMemoryUploadedFile(image_io, None, name, None, None, None)

    def create(self, data):
        try:
            model = self.__class__.Meta.model
        except () as e:
            raise ValueError(e)
        else:
            query = model.objects.filter(title__iexact=data['title'])
        if query.exists():
            return None
        else:
            self.data_db = model.objects.create(**data)
            return self.data_db

    def is_field_coincidence(self):
        if hasattr(self.__class__, 'Meta') and hasattr(self.__class__.Meta, 'field_coincidence'):
            return True
        return False

    def log_output(self, result):
        logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)
        logging.info('{} :\n {}'.format(self, result))

    @staticmethod
    def __get_result(it):
        pars_result = dict()
        key_ind = int()
        for res in it:
            if isinstance(res, int):
                pars_result[res] = dict()
                key_ind = res
            else:
                pars_result[key_ind].update(res)
        return pars_result

    def run(self, log=False, create=True):
        parser = ResultParser(self)
        pars_result = self.__get_result(parser)
        if not create and not log:
            return pars_result
        for _, out_data in pars_result.items():
            if log:
                self.log_output(out_data)
            elif create and out_data:
                out_data.update(self.add_field)
                self.create(out_data)




import os
from setuptools import setup
from djparsing import __author__, __version__


def __read_requires(file):
    try:
        return open(os.path.join(os.path.dirname(__file__), file)).read()
    except IOError:
        return ''


install_requires = __read_requires("requirements.txt").split()

setup(
    name='djparsing',
    description='microframework for parsing sites, '
                'a simple interface and flexibility will help you quickly start parsing sites.',
    version=__version__,
    packages=['djparsing'],
    author=__author__,
    author_email='vlinchevskyi@gmail.com',
    license='BSD License',
    install_requires=install_requires,
    url='https://github.com/lincvv/djparsing',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities',
    ],
)
